"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const readline = require("readline");
const serendipity_public_1 = require("@attachments/serendipity-public");
const ErrorStackParser = require("error-stack-parser");
const boxen = require("boxen");
const utils_1 = require("./utils");
const Table = require('cli-table');
class SerendipityWebpackPlugin {
    serverPort;
    serverHost;
    analysisPort;
    constructor(option) {
        this.serverPort = option.webpackDevServerPort;
        this.analysisPort = option.webpackAnalysisPort;
        this.serverHost = option.webpackDevServerHost;
    }
    static PLUGIN_INFO = {
        name: 'SerendipityWebpackPlugin'
    };
    clearConsole() {
        readline.cursorTo(process.stdout, 0, 0);
        readline.clearScreenDown(process.stdout);
    }
    resolveProblems(stats, type) {
        const findErrorsRecursive = (compilation) => {
            const errors = compilation[type];
            if (errors.length === 0 && compilation.children) {
                for (const child of compilation.children) {
                    errors.push(...findErrorsRecursive(child));
                }
            }
            return utils_1.uniqueBy(errors, error => error.message);
        };
        return findErrorsRecursive(stats.compilation);
    }
    showProblems(problems, type) {
        serendipity_public_1.logger.info('o(TωT)o 请解决下面的 bug:\n');
        const errors = problems.map(res => {
            return {
                message: res.message,
                name: res.name,
                webpackError: res,
                errorStack: res.stack ? ErrorStackParser.parse(res) : []
            };
        });
        errors.forEach(res => {
            type === 'warnings' ? serendipity_public_1.logger.warn(res.name + ':') : serendipity_public_1.logger.error(res.name + ':');
            console.log(`${res.message}\n\n`);
        });
    }
    onWebpackDone(stats) {
        this.clearConsole();
        const hasError = stats.hasErrors();
        const hasWarnings = stats.hasWarnings();
        if (!hasError && !hasWarnings) {
            this.onWebpackSuccess(stats);
            return;
        }
        if (hasError) {
            const problems = this.resolveProblems(stats, 'errors');
            this.showProblems(problems, 'errors');
            return;
        }
        if (hasWarnings) {
            const problems = this.resolveProblems(stats, 'warnings');
            this.showProblems(problems, 'warnings');
            return;
        }
    }
    onWebpackInvalid() {
        this.clearConsole();
        serendipity_public_1.logger.info('(〃\'▽\'〃) 项目正在构建中，请稍候...');
    }
    onWebpackSuccess(stats) {
        const timeCost = stats.endTime - stats.startTime;
        serendipity_public_1.logger.done('Compile successfully!');
        console.log(boxen(`ヾ(๑╹◡╹)ﾉ" 项目构建成功! (耗时：${timeCost} ms)\n\n项目运行地址：http://127.0.0.1:${this.serverPort}\n查看 webpack 打包分析：http://127.0.0.1:${this.analysisPort}`, { padding: 1 }));
        const table = new Table({
            head: ['文件名', '大小']
        });
        const chunksInfo = (stats.toJson().chunks);
        table.push(...utils_1.getChunkInfo(chunksInfo, 10));
        console.log('');
        serendipity_public_1.logger.info('资源清单如下：');
        console.log(table.toString());
    }
    apply(compiler) {
        if (compiler.hooks) {
            compiler.hooks.done.tap(SerendipityWebpackPlugin.PLUGIN_INFO, this.onWebpackDone.bind(this));
            compiler.hooks.invalid.tap(SerendipityWebpackPlugin.PLUGIN_INFO, this.onWebpackInvalid.bind(this));
        }
    }
}
exports.default = SerendipityWebpackPlugin;
//# sourceMappingURL=serendipityWebpackPlugin.js.map